# -*- coding: utf-8 -*-

from django.apps import apps
from django.conf import settings
from django.db import connections, transaction


class HistoryHelper:
    """Helper class that provides the `sync_history_tables` method."""

    def __init__(self):
        self.info_list = []
        self.success_list = []
        self.error_list = []
        self.success_message = None
        self.error_message = None

    def sync_history_tables(self, batch_size=100):
        """
        Sync data from default database to history database.
        This operation is done using atomic transactions for both databases.
        """
        if settings.HISTORY_DATABASE_NAME == "default":
            self.info_list.append(
                "Skipping history sync because the setting HISTORY_DATABASE_NAME is the default database."
            )
            return self._build_result()

        model_classes = self._get_pghistory_model_classes()
        self._sync_history_tables(model_classes, batch_size, pk_key="id")

        model_classes = self._get_history_model_classes()
        self._sync_history_tables(model_classes, batch_size, pk_key="pgh_id")

        return self._build_result()

    def _sync_history_tables(self, model_classes, batch_size=10, pk_key="pgh_id"):
        self.info_list.append(f"History table to sync {len(model_classes)}")
        for model_class in model_classes:
            db_table_name = model_class._meta.db_table

            while True:
                try:
                    queryset = model_class.objects.using("default").all().order_by(pk_key)[:batch_size]
                    count = queryset.only(pk_key).count()

                    if count == 0:
                        break

                    self.info_list.append(f"History to sync for {db_table_name}: {count}")
                    with transaction.atomic():
                        with transaction.atomic(using=settings.HISTORY_DATABASE_NAME):
                            values = list(queryset.values())  # evaluate data

                            # Bulk insert on history db & set sequence
                            (
                                model_class.objects
                                .using(settings.HISTORY_DATABASE_NAME)
                                .bulk_create([model_class(**vals) for vals in values])
                            )
                            # Bulk delete on default db
                            pks = [v[pk_key] for v in values]
                            model_class.objects.using("default").filter(**{f"{pk_key}__in": pks}).delete()

                            self.success_list.extend([f"History for {db_table_name}, pk {pk} saved" for pk in pks])
                except Exception as emc:
                    self.error_list.append(
                        {f'Error sync history for all {db_table_name}': str(emc)}
                    )
                    break

    def _build_result(self):
        return {
            'info_list': self.info_list,
            'success_list': self.success_list,
            'error_list': self.error_list,
            'success_message': self.success_message,
            'error_message': self.error_message
        }

    @staticmethod
    def _get_history_model_classes():
        return [x for x in apps.get_models() if x._meta.app_label == settings.HISTORY_APP_NAME]

    @staticmethod
    def _get_pghistory_model_classes():
        return [apps.get_model("pghistory", "Context")]
